"""Describe what tasks this Process accomplishes."""
# Order of imported packages should be:
# Standard libraries
from pathlib import Path
from shutil import copy2

# Third party libraries
from plumbum import TEE

# Resolwe
from resolwe.process import (
    Process,
    SchedulingClass,
    Persistence,
    Cmd,
    StringField,
    TextField,
    BooleanField,
    IntegerField,
    FloatField,
    DateField,
    DateTimeField,
    UrlField,
    SecretField,
    FileField,
    FileHtmlField,
    DirField,
    JsonField,
    ListField,
    DataField,
    GroupField,
)

# In-house libraries (i.e. resolwe_bio)
# Local libraries


class YourProcessName(Process):
    """Process description in one line ends with a period.

    Continue with a more elaborate description of what the process
    does, short layout of the algorithm and what output(s) to expect.

    Note that this is not a working process, but rather just shows
    bits and pieces that can be used in your real process.
    """

    slug = "process-slug-name"
    name = "Human readable name of the process"
    process_type = "data:text_processing:"
    version = "1.0.0"
    # A category of the process, can be used in i.e. frontend app.
    category = "Read Processing"
    # Dynamically generated name of the resulting datum object. Lookup
    # happens from Input class fields.
    data_name = 'Data object ({{ string_field.file|default("?") }})'
    scheduling_class = SchedulingClass.BATCH  # or .INTERACTIVE for short jobs
    # or RAW or TMP, see documentation at
    # https://resolwe.readthedocs.io/en/latest/proc.html#persistence
    persistence = Persistence.CACHED
    # Use below entity to make the created object also an entity of
    # type sample. See
    # https://resolwe.readthedocs.io/en/latest/proc.html#entity for
    # more details.
    # Fields descriptor_schema, input and always_create are optional.
    entity = {
        "type": "sample",  # name of your entity
        # Use a descriptor schema slug if it differs from "type".
        "descriptor_schema": "descriptor-schema-of-entity",
        "input": "name of input that has the entity",
        # If you want to create entity in all cases
        "always_create": True,
    }
    requirements = {
        "expression-engine": "jinja",
        "executor": {
            "docker": {
                # Specify Docker image, may need to include a version.
                "image": "resolwe/base:fedora-31"
            }
        },
        "resources": {
            "cores": 8,
            "memory": 16384,
            # Connects container to a network.
            "network": True,
        },
    }

    class Input:
        """Input fields to process MergeData."""

        string_field = StringField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
            choices=[
                ("computer_readable1", "Human readable 1"),
                ("computer_readable2", "Human readable 2"),
            ],
            default="computer_readable1",
            required=False,
            hidden=False,
            allow_custom_choice=True,
        )
        text_field = TextField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
            default="default text",
            required=False,
            hidden=True,
        )
        boolean_field1 = BooleanField(
            label="Labels are short and do not end in a period",
            description="Note that description fields always end in a period.",
            default=False,
            required=True,
            hidden=False,
        )
        integer_field = IntegerField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
            default=1,
        )
        float_field = FloatField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
            default=3.14,
        )
        date_field = DateField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
            default="2020-04-20",
        )
        datetime_field = DateTimeField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
            default="2020-04-20 12:16:00",
        )
        url_field = UrlField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
        )
        secret_field = SecretField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
        )
        file_field = FileField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
        )
        filehtml_field = FileHtmlField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
        )
        dir_field = DirField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
        )
        json_field = JsonField(
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
        )
        list_field = ListField(
            DataField(data_type="your:data:type"),
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
        )
        data_field = DataField(
            # data_type should not start with data:
            data_type="your:data:type",
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
        )

        class Advanced:
            """Add advanced list of options."""

            boolean_field2 = BooleanField(
                label="Labels are short and do not end in a period",
                description="Description ends in a period.",
                default=False,
            )

        group_field = GroupField(
            Advanced,
            label="Labels are short and do not end in a period",
            disabled=False,
            # Will show when boolean_field1 is flipped.
            hidden="!boolean_field1",
            collapsed=True,
        )

    class Output:
        """Output fields to process MergeData"""

        file_out = FileField(label="Labels are short and do not end in a period")
        dir_optional = DirField(
            label="Labels are short and do not end in a period", required=False
        )
        list_out = ListField(
            FileField(),
            label="Labels are short and do not end in a period",
            description="Description ends in a period.",
        )

    def run(self, inputs, outputs):
        """Enter your not so pseudo code below

        Inputs are available through `inputs` variable. I.e., using
        `inputs.list_fields` should yield a list.

        """

        # Below code extracts file name and omits .txt extension.
        # Assuming Input has a variable called file_field, which in
        # turn has field raw_file which holds a path to the file.
        basename = Path(inputs.file_field.raw_file.path).name
        assert basename.endswith(".txt")
        name = basename[:-4]

        # Possible values for various fields.
        inputs.string_field  # string ('computer_readable1')
        inputs.text_field  # string ('default text')
        inputs.boolean_field  # boolean True or False
        inputs.integer_field  # integer scalar (1)
        inputs.float_field  # float scalar (3.14)
        inputs.date_field  # `yyy-mm-dd` (2020-04-20)
        inputs.datetime_field  # `yyy-mm-dd hh:mm:ss` (2020-04-20 12:16:00)
        inputs.secret_field  # TODO: secret model
        inputs.file_field  # string (relative) path to file
        inputs.filehtml_field  # string (relative) path to file
        inputs.dir_field  # string (relative) path to directory
        inputs.json_field  # string (relative) path to file
        inputs.list_field  # list [of, subordinate, fields]
        # Possible self-explanatory properties of DataField.
        inputs.data_field.id
        inputs.data_field.type
        inputs.data_field.descriptor
        inputs.data_field.entity_name
        inputs.url_field  # string or dict {'file': '', refs: ''}
        # You can access fields in GroupField entries from Input.
        inputs.group_field.boolean_field2

        # Run a simple command line tool.
        (Cmd["ls"]["-lha"] > "output_file.txt")()

        # You can use piping as in a terminal.
        (Cmd["ls"] | Cmd["wc"]["-l"])()

        # To run a command line tool and capture its output, use TEE.
        args = ["-l", "-a", "-h"]
        return_code, std_out, std_err = Cmd["ls"][args] & TEE(
            retcode=None,
            buffered=True,
            timeout=None
            # returns None if process exists sucessfully
        )

        # Report errors, warning or info using:
        self.error(f"Oops, an error occurred in {inputs.data_field.id}")
        self.warning(f"Warning in {inputs.data_field.id}")
        self.info(f"Pass info about {self.version}")

        # You can move the progress bar using a float on scale 0.0 to 1.0.
        self.progress(0.5)

        # Specify all fields in Output class to `outputs`.
        outputs.file_out = "./path/to/filename.txt"
        outputs.list_out = ["./path/to/file1.txt", "./path/to/file2.txt"]

        # To import file from the input, you can use import_file() function.
        imported_file = inputs.file_field.import_file()

        # You need to export all fields except those that are explicitly
        # set to required=False.
        if Path("./result_folder").exists:
            outputs.dir_optional = "./result_folder"



# Generated by Django 3.2.12 on 2022-10-26 12:41

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0006_add_descriptor_to_relation"),
    ]

    operations = [
        migrations.AlterField(
            model_name="data",
            name="process_rc",
            field=models.SmallIntegerField(blank=True, null=True),
        ),
    ]



# pylint: disable=missing-docstring
import tempfile
from pathlib import Path
from typing import Sequence

from resolwe.storage.connectors import utils
from resolwe.test import TestCase


def create_file(file: Path, size: int):
    with file.open("wb") as f:
        f.seek(size - 1)
        f.write(b"\0")
        f.seek(0)
        f.write(file.name.encode("utf-8"))


class UtilsTest(TestCase):
    def test_get_transfer_object_file(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            file = tmp_path / "file.in"
            create_file(file, 1000)
            to_transfer = utils.get_transfer_object(file, tmp_path)
            self.assertDictEqual(
                to_transfer,
                {
                    "size": 1000,
                    "path": "file.in",
                    "md5": "9638a76565e9f1f15443421c5605d8ce",
                    "crc32c": "5e29c10c",
                    "awss3etag": "9638a76565e9f1f15443421c5605d8ce",
                    "chunk_size": 8 * 1024 * 1024,
                },
            )

    def test_get_transfer_object_directory(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            directory: Path = tmp_path / "dir"
            directory.mkdir()
            to_transfer = utils.get_transfer_object(directory, tmp_path)
            self.assertDictEqual(
                to_transfer,
                {
                    "size": 0,
                    "path": "dir/",
                    "md5": "",
                    "crc32c": "",
                    "awss3etag": "",
                },
            )

    def test_get_transfer_object_nonexisting(self):
        to_transfer = utils.get_transfer_object("base/nonexisting.txt", "base")
        self.assertIsNone(to_transfer)

    def test_chunks(self):
        parts = range(100)
        chunks = list(utils.chunks(parts, 4))
        self.assertEqual(len(chunks), 4)
        for chunk in chunks:
            self.assertEqual(len(chunk), 25)

        chunks = list(utils.chunks(parts, 200))
        self.assertEqual(len(chunks), 200)
        for chunk in chunks[: len(parts)]:
            self.assertEqual(len(chunk), 1)
        for chunk in chunks[len(parts) :]:
            self.assertEqual(len(chunk), 0)

        chunks = list(utils.chunks(parts, 3))
        self.assertEqual(len(chunks), 3)
        self.assertEqual(len(chunks[0]), 34)
        self.assertEqual(len(chunks[1]), 33)
        self.assertEqual(len(chunks[2]), 33)

    def test_paralelize(self):
        parts = [1] * 100

        def worker(parts: Sequence[int]) -> int:
            return sum(parts)

        def worker_exception(parts: Sequence[int]) -> int:
            if len(parts) == 34:
                raise Exception()
            return sum(parts)

        futures = utils.paralelize(parts, worker)
        results = [f.result() for f in futures]
        self.assertEqual(results, [10] * 10)

        futures = utils.paralelize(parts, worker, max_threads=200)
        results = [f.result() for f in futures]
        self.assertEqual(results, [1] * 100)

        futures = utils.paralelize(parts, worker, max_threads=3)
        results = [f.result() for f in futures]
        self.assertEqual(results, [34, 33, 33])

        futures = utils.paralelize(parts, worker_exception, max_threads=3)
        with self.assertRaises(Exception):
            futures[0].result()
        results = [f.result() for f in futures[1:]]
        self.assertEqual(results, [33, 33])



""".. Ignore pydocstyle D400.

==================
List Docker images
==================

"""

import functools
import logging
import operator
import shlex
import subprocess
import threading
import time

import yaml

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from resolwe.flow.models import Process

logger = logging.getLogger(__name__)

# Global list of images already pulled when running in the same process. This
# is used to avoid pulling images more than once when executed from tests.
PULLED_IMAGES = set()
PULLED_IMAGES_LOCK = threading.Lock()


class Command(BaseCommand):
    """List Docker images used by processes.  Optionally also pull them."""

    help = "List Docker images used by processes in either plain text or YAML (for Ansible)"

    def add_arguments(self, parser):
        """Add an argument to specify output format."""
        parser.add_argument(
            "--format",
            dest="format",
            default="plain",
            help="Set output format ('plain' [default] or 'yaml')",
        )
        parser.add_argument(
            "--pull",
            dest="pull",
            default=False,
            action="store_true",
            help="Pull all images with Docker",
        )
        parser.add_argument(
            "--ignore-pull-errors",
            dest="ignore_pull_errors",
            default=getattr(settings, "FLOW_DOCKER_IGNORE_PULL_ERRORS", False),
            action="store_true",
            help="Don't fail whenever a Docker image can't be pulled",
        )

    def handle(self, *args, **options):
        """Handle command list_docker_images."""
        verbosity = int(options.get("verbosity"))

        # Check that the specified output format is valid
        if options["format"] != "plain" and options["format"] != "yaml":
            raise CommandError("Unknown output format: %s" % options["format"])

        # Gather only unique latest custom Docker requirements that the processes are using
        # The 'image' field is optional, so be careful about that as well
        unique_docker_images = set(
            p.requirements["executor"]["docker"]["image"]
            for p in Process.objects.filter(is_active=True)
            .order_by("slug", "-version")
            .distinct("slug")
            .only("requirements")
            .filter(requirements__icontains="docker")
            if "image" in p.requirements.get("executor", {}).get("docker", {})
        )

        # Add the default image.
        unique_docker_images.add(
            getattr(
                settings,
                "FLOW_DOCKER_DEFAULT_PROCESSING_CONTAINER_IMAGE",
                "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-20.04",
            )
        )

        # Pull images if requested or just output the list in specified format
        if options["pull"]:
            # Remove set of already pulled images.
            with PULLED_IMAGES_LOCK:
                unique_docker_images.difference_update(PULLED_IMAGES)

            # Get the desired 'docker' command from settings or use the default
            docker = getattr(settings, "FLOW_DOCKER_COMMAND", "docker")

            # Pull each image
            for img in unique_docker_images:
                max_retries = 10
                for _ in range(max_retries):
                    ret = subprocess.run(
                        shlex.split("{} pull {}".format(docker, img)),
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                    )
                    if ret.returncode == 0:
                        break
                    time.sleep(1)

                # Update set of pulled images.
                with PULLED_IMAGES_LOCK:
                    PULLED_IMAGES.add(img)

                if ret.returncode != 0:
                    errmsg = "Failed to pull Docker image '{}': '{}'.".format(
                        img, ret.stdout.decode()
                    )

                    if not options["ignore_pull_errors"]:
                        # Print error and stop execution
                        raise CommandError(errmsg)
                    else:
                        # Print error, but keep going
                        logger.error(errmsg)
                        if verbosity > 0:
                            self.stderr.write(errmsg)
                else:
                    msg = "Docker image '{}' pulled successfully!".format(img)
                    logger.info(msg)
                    if verbosity > 0:
                        self.stdout.write(msg)
        else:
            # Sort the set of unique Docker images for nicer output.
            unique_docker_images = sorted(unique_docker_images)

            # Convert the set of unique Docker images into a list of dicts for easier output
            imgs = [
                dict(name=s[0], tag=s[1] if len(s) == 2 else "latest")
                for s in (img.split(":") for img in unique_docker_images)
            ]

            # Output in YAML or plaintext (one image per line), as requested
            if options["format"] == "yaml":
                out = yaml.safe_dump(imgs, default_flow_style=True, default_style="'")
            else:
                out = functools.reduce(
                    operator.add, ("{name}:{tag}\n".format(**i) for i in imgs), ""
                )

            self.stdout.write(out, ending="")



"""Consumers for Observers."""

import asyncio
from enum import Enum
from typing import Callable, Optional

from channels.consumer import AsyncConsumer
from channels.db import database_sync_to_async
from channels.generic.websocket import JsonWebsocketConsumer

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.utils import timezone

from resolwe.flow.models.utils import bulk_duplicate

from .models import BackgroundTask, Observer, Subscription
from .protocol import GROUP_SESSIONS, ChangeType, ChannelsMessage, WebsocketMessage

# The channel used to listen for BackgrountTask events
BACKGROUND_TASK_CHANNEL = "observers.background_task"


class BackgroundTaskType(Enum):
    """Background task types."""

    DUPLICATE_DATA = "duplicate_data"
    DUPLICATE_ENTITY = "duplicate_entity"
    DUPLICATE_COLLECTION = "duplicate_collection"
    DELETE = "delete"


def update_constants():
    """Recreate channel name constant with changed settings.

    This kludge is mostly needed due to the way Django settings are
    patched for testing and how modules need to be imported throughout
    the project. On import time, settings are not patched yet, but some
    of the code needs static values immediately. Updating functions such
    as this one are then needed to fix dummy values.
    """
    global BACKGROUND_TASK_CHANNEL
    redis_prefix = getattr(settings, "FLOW_MANAGER", {}).get("REDIS_PREFIX", "")

    BACKGROUND_TASK_CHANNEL = f"{redis_prefix}.observers.background_task"


update_constants()


def database_sync_to_async_new_thread(*args, **kwargs):
    """Patched database_sync_to_async decorator.

    That sets thread_sensitive to True by default, otherwise the defalut executor tends
    to lock.
    """
    target_arg_name = "thread_sensitive"
    # Make sure not to override explicitely set flag.
    if target_arg_name not in kwargs:
        kwargs[target_arg_name] = False
    return database_sync_to_async(*args, **kwargs)


class ClientConsumer(JsonWebsocketConsumer):
    """Consumer for client communication."""

    def websocket_connect(self, event: dict[str, str]):
        """Handle establishing a WebSocket connection."""
        session_id: str = self.scope["url_route"]["kwargs"]["session_id"]
        self.session_id = session_id

        # Accept the connection.
        super().websocket_connect(event)

    @property
    def groups(self) -> list[str]:
        """Generate a list of groups this channel should add itself to."""
        if not hasattr(self, "session_id"):
            return []
        return [GROUP_SESSIONS.format(session_id=self.session_id)]

    def disconnect(self, code: int):
        """Handle closing the WebSocket connection."""
        Subscription.objects.filter(session_id=self.session_id).delete()

    def observers_item_update(self, msg: ChannelsMessage):
        """Handle an item update signal."""
        content_type = ContentType.objects.get_for_id(msg["content_type_pk"])
        object_id = msg["object_id"]
        change_type = ChangeType(msg["change_type_value"])
        source = msg["source"]

        interested = Observer.get_interested(
            content_type=content_type, object_id=object_id, change_type=change_type
        )
        subscription_ids = [
            subscription_id.hex
            for subscription_id in Subscription.objects.filter(observers__in=interested)
            .filter(session_id=self.session_id)
            .values_list("subscription_id", flat=True)
            .distinct()
        ]

        is_object_source = source == (content_type.name, object_id)
        if change_type == ChangeType.DELETE and is_object_source:
            # The observed object was either deleted or the user lost permissions.
            subscription = Subscription.objects.get(session_id=self.session_id)
            observers = Observer.objects.filter(
                content_type=content_type,
                object_id=object_id,
            )
            # Assure we don't stay subscribed to an illegal object.
            subscription.observers.remove(*observers)

        to_send: WebsocketMessage = {
            "object_id": object_id,
            "change_type": change_type.name,
            "subscription_id": "0",
            "source": source,
        }

        for subscription_id in subscription_ids:
            to_send["subscription_id"] = subscription_id
            self.send_json(to_send)


class BackgroundTaskConsumer(AsyncConsumer):
    """The background task consumer."""

    @database_sync_to_async_new_thread
    def wrap_task(self, function: Callable, task_id: int):
        """Start the function and update background task status."""
        task: Optional[BackgroundTask] = None
        try:
            task = BackgroundTask.objects.get(pk=task_id)
            task.started = timezone.now()
            task.status = BackgroundTask.STATUS_PROCESSING
            task.save(update_fields=["status", "started"])
            task.output = function() or "Task completed."
            task.status = BackgroundTask.STATUS_DONE
        # Task may not exist here if the consumer was cancelled duging the creation
        # of the task. Only proceed if it is defined.
        except ValidationError as e:
            if task:
                task.status = BackgroundTask.STATUS_ERROR
                task.output = e.messages
        except asyncio.CancelledError:
            if task:
                task.status = BackgroundTask.STATUS_ERROR
                task.output = "Task was cancelled."
        except Exception as e:
            if task:
                task.status = BackgroundTask.STATUS_ERROR
                task.output = str(e)
        finally:
            if task:
                task.finished = timezone.now()
                task.save(update_fields=["status", "finished", "output"])

    async def duplicate_data(self, message: dict):
        """Duplicate the data and update task status."""
        # Break circular import.
        from resolwe.flow.models import Data

        def duplicate():
            duplicates = bulk_duplicate(
                data=Data.objects.filter(pk__in=message["data_ids"]),
                contributor=get_user_model().objects.get(pk=message["contributor_id"]),
                inherit_entity=message["inherit_entity"],
                inherit_collection=message["inherit_collection"],
            )
            return list(duplicates.values_list("pk", flat=True))

        await self.wrap_task(duplicate, message["task_id"])

    async def duplicate_entity(self, message: dict):
        """Duplicate the entities and update task status."""
        # Break circular import.
        from resolwe.flow.models import Entity

        def duplicate():
            duplicates = bulk_duplicate(
                entities=Entity.objects.filter(pk__in=message["entity_ids"]),
                contributor=get_user_model().objects.get(pk=message["contributor_id"]),
                inherit_collection=message["inherit_collection"],
            )
            return list(duplicates.values_list("pk", flat=True))

        await self.wrap_task(duplicate, message["task_id"])

    async def duplicate_collection(self, message: dict):
        """Duplicate the collections and update task status."""
        # Break circular import.
        from resolwe.flow.models import Collection

        def duplicate():
            duplicates = bulk_duplicate(
                collections=Collection.objects.filter(pk__in=message["collection_ids"]),
                contributor=get_user_model().objects.get(pk=message["contributor_id"]),
            )
            return list(duplicates.values_list("pk", flat=True))

        await self.wrap_task(duplicate, message["task_id"])

    async def delete(self, message: dict):
        """Delete the objects and update task status.

        :param message: the message containing the following keys:
            - ``task_id``: the id of the background task
            - ``content_type_id``: the content type of the model to be deleted
            - ``object_ids``: the ids of the objects to be deleted
        """

        def delete():
            Model = ContentType.objects.get_for_id(
                message["content_type_id"]
            ).model_class()
            to_delete = Model.objects.filter(pk__in=message["object_ids"])

            # Use iterator and delete one object at a time to trigger the signals.
            for obj in to_delete.iterator():
                obj.delete()

        await self.wrap_task(delete, message["task_id"])



"""Resolwe annotations model."""
import abc
import datetime
import re
from collections import defaultdict
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    MutableSequence,
    Optional,
    Sequence,
    Type,
)

from django.core.exceptions import ValidationError
from django.db import models

from resolwe.permissions.models import PermissionObject

if TYPE_CHECKING:
    from resolwe.flow.models import Entity, Collection

from .base import AuditModel


class HandleMissingAnnotations(Enum):
    """How to handle missing annotations."""

    ADD = "ADD"
    REMOVE = "REMOVE"


class AnnotationType(Enum):
    """Supported annotation types."""

    DATE = "DATE"
    DECIMAL = "DECIMAL"
    INTEGER = "INTEGER"
    STRING = "STRING"


class AnnotationValueValidator:
    """Class that validates the annotation value.

    It uses plugins (classes) to validate the annotation of the given type.
    """

    def __init__(self):
        """Initialize."""
        self._validators: Mapping[
            AnnotationType, MutableSequence["AnnotationFieldBaseValidator"]
        ] = defaultdict(list)

    def _add_validator(self, validator: "AnnotationFieldBaseValidator"):
        """Add the given validator."""
        for field_type in validator.annotation_field_types:
            self._validators[field_type].append(validator)

    def validate(
        self, annotation_value: "AnnotationValue", raise_exception: bool = False
    ) -> Sequence[str]:
        """Validate the given AnnotationValue object.

        All validation errors are grouped together.

        :return: sequence of validation errors.
        :raises ValidationError: if annotation fails and raise_exception is True. The
            actual errors are nested inside outer ValidationError.
        """
        errors = []
        validators = self._validators[annotation_value.field.annotation_type]
        for validator in validators:
            try:
                validator.validate(annotation_value.value, annotation_value.field)
            except ValidationError as error:
                errors.append(error)
        if errors and raise_exception:
            raise ValidationError(errors)
        return errors


annotation_value_validator = AnnotationValueValidator()


class AnnotationFieldBaseValidator(metaclass=abc.ABCMeta):
    """Base class for annotation class validation."""

    annotation_field_types: Iterable[AnnotationType]

    @classmethod
    def __init_subclass__(cls: Type["AnnotationFieldBaseValidator"], **kwargs):
        """Register instance of the class with the validator."""
        super().__init_subclass__(**kwargs)
        annotation_value_validator._add_validator(cls())

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate the value.

        :raises ValidatiorError: on validation failure.
        """
        raise NotImplementedError("Subclass should implement the 'validate' method.")


class AnnotationFieldStringValidator(AnnotationFieldBaseValidator):
    """Validates the string field."""

    annotation_field_types = (AnnotationType.STRING,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value to be of type str."""
        if not isinstance(value, str):
            raise ValidationError(
                f"The value '{value}' is not of the expected type 'str'."
            )


class AnnotationFieldIntegerValidator(AnnotationFieldBaseValidator):
    """Validates the integer field."""

    annotation_field_types = (AnnotationType.INTEGER,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value to be of type int."""
        if not isinstance(value, int):
            raise ValidationError(
                f"The value '{value}' is not of the expected type 'int'."
            )


class AnnotationFieldDecimalValidator(AnnotationFieldBaseValidator):
    """Validates the decimal field."""

    annotation_field_types = (AnnotationType.DECIMAL,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value to be of type int or float."""
        if not isinstance(value, (int, float)):
            raise ValidationError(
                f"The value '{value}' is not of the expected types 'int' or 'float'."
            )


class AnnotationFieldDateValidator(AnnotationFieldBaseValidator):
    """Validates the date field."""

    annotation_field_types = (AnnotationType.DATE,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value represents a date in the format YYYY-MM-DD.

        The month and day part can contain a single character.
        """
        if not isinstance(value, str):
            raise ValidationError(
                f"The value '{value}' is not of the expected type 'str'."
            )
        try:
            datetime.datetime.strptime(value, "%Y-%m-%d")
        except ValueError:
            raise ValidationError(
                f"Value {value} has incorrect format, use YYYY-MM-DD."
            )


class AnnotationFieldRegexValidator(AnnotationFieldBaseValidator):
    """Validates regex for the annotation value."""

    annotation_field_types = (
        AnnotationType.INTEGER,
        AnnotationType.DATE,
        AnnotationType.DECIMAL,
        AnnotationType.STRING,
    )

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate that the value matches regex defined on the field.

        The value is converted to string before matching.
        """
        # Validate the regex.
        if annotation_field.validator_regex is not None:
            regex = annotation_field.validator_regex
            value = str(value)
            if re.search(regex, value) is None:
                raise ValidationError(
                    f"The value '{value}' for the field '{annotation_field.pk}' does "
                    f"not match the regex '{regex}'."
                )


class AnnotationFieldVocabularyValidator(AnnotationFieldBaseValidator):
    """Validates that the value matches vocabulary."""

    annotation_field_types = (
        AnnotationType.INTEGER,
        AnnotationType.DATE,
        AnnotationType.DECIMAL,
        AnnotationType.STRING,
    )

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate that the value matches vocabulary on the field."""
        if annotation_field.vocabulary and value not in annotation_field.vocabulary:
            raise ValidationError(
                f"The value '{value}' is not valid for the field {annotation_field}."
            )


class AnnotationGroup(models.Model):
    """Group of annotation fields."""

    #: the name of the annotation group
    name = models.CharField(max_length=128)

    #: the label of the annotation group
    label = models.CharField(max_length=128)

    #: the sorting order among annotation groups
    sort_order = models.PositiveSmallIntegerField()

    def __str__(self) -> str:
        """Return user-friendly string representation."""
        return f"{self.name}"

    class Meta:
        """Set the default ordering."""

        ordering = ["sort_order"]


class AnnotationField(models.Model):
    """Annotation field."""

    #: the name of the annotation fields
    name = models.CharField(max_length=128)

    #: user visible field name
    label = models.CharField(max_length=128)

    #: user visible field description
    description = models.CharField(max_length=256)

    #: the type of the annotation field
    type = models.CharField(max_length=16)

    #: the annotation group this field belongs to
    group = models.ForeignKey(
        AnnotationGroup, on_delete=models.CASCADE, related_name="fields"
    )

    #: the sorting order among annotation fields
    sort_order = models.PositiveSmallIntegerField()

    #: optional regular expression for validation
    validator_regex = models.CharField(max_length=128, null=True)

    #: optional map of valid values to labels
    vocabulary = models.JSONField(null=True)

    def label_by_value(self, label: str) -> str:
        """Get the value by label.

        When no value is found the label is returned.
        """
        return {label: value for value, label in self.vocabulary.items()}.get(
            label, label
        )

    @staticmethod
    def add_to_collection(source: "Collection", destination: "Collection"):
        """Add fields from the source to the destination collection."""
        destination.annotation_fields.add(*source.annotation_fields.all())

    @property
    def annotation_type(self) -> AnnotationType:
        """Get the field type as enum."""
        return AnnotationType(self.type.upper())

    @staticmethod
    def group_field_from_path(path: str) -> List[str]:
        """Return the group and field name from path."""
        return path.split(".", maxsplit=1)

    @classmethod
    def id_from_path(cls, path: str) -> Optional[int]:
        """Get the field id from the field path."""
        group_name, field_name = cls.group_field_from_path(path)
        return (
            cls.objects.filter(group__name=group_name, name=field_name)
            .values_list("id", flat=True)
            .first()
        )

    @classmethod
    def field_from_path(cls, path: str) -> Optional["AnnotationField"]:
        """Get the field id from the field path."""
        group_name, field_name = cls.group_field_from_path(path)
        return cls.objects.filter(group__name=group_name, name=field_name).first()

    def __str__(self) -> str:
        """Return user-friendly string representation."""
        return f"{self.group.name}.{self.name}"

    class Meta:
        """Set the constraints and the default ordering."""

        constraints = [
            # Accept only supported annotation types.
            models.constraints.CheckConstraint(
                check=models.Q(type__in=[e.value for e in AnnotationType]),
                name="annotation_type",
            ),
            models.constraints.UniqueConstraint(
                fields=["name", "group"], name="uniquetogether_name_group"
            ),
        ]
        ordering = ["group__sort_order", "sort_order"]


class AnnotationPreset(AuditModel, PermissionObject):
    """The named set of annotation fields.

    The presets have permissions.
    """

    #: the name of the annotation fields
    name = models.CharField(max_length=128)

    #: the fields belongint to this preset
    fields = models.ManyToManyField(AnnotationField, related_name="presets")

    class Meta:
        """Override parent meta."""


class AnnotationValue(AuditModel):
    """The value of the annotation."""

    #: the entity this field belongs to
    entity: "Entity" = models.ForeignKey(
        "Entity", related_name="annotations", on_delete=models.CASCADE
    )

    #: the field this field belongs to
    field: AnnotationField = models.ForeignKey(
        AnnotationField, related_name="values", on_delete=models.CASCADE
    )

    #: value is stored under key 'value' in the json field to simplify lookups
    _value: Any = models.JSONField(default=dict)

    def __init__(self, *args, **kwargs):
        """Allow us to set the 'value' in the constructor.

        This is a little tricky, as value can be passed in args or kwargs. We only
        intercept the value when given as kwarg, since args are used by Django when
        constructing class from database value.
        """
        if "value" in kwargs:
            kwargs["_value"] = {"value": kwargs.pop("value")}
        super().__init__(*args, **kwargs)

    @property
    def value(self):
        """Get the actual value."""
        return self._value["value"]

    @value.setter
    def value(self, value):
        """Set the value.

        The object must be saved for the value to propagate to the database.
        """
        self._value["value"] = value

    def validate(self):
        """Validate the given value.

        The validation is always done in full, all errors are gathered and sent
        in the response.

        :raises ValidationError: when the validation fails.
        """
        annotation_value_validator.validate(self, raise_exception=True)

    @property
    def value_label_map(self) -> Dict[str, Any]:
        """Compute the mapping from value to label.

        Only applicable when vocabulary is given, otherwise empty dictionary is
        returned.
        """
        return self.field.vocabulary or {}

    @property
    def label(self) -> Any:
        """Get the value of the label from value.

        This is only applicable if the vocabulary is set. When no vocabulary is
        set it returns the value.
        """
        return self.field.vocabulary.get(self.value, self.value)

    def __str__(self) -> str:
        """Return user-friendly string representation."""
        return f"{self.value}"



"""Process viewset."""
from rest_framework import exceptions, mixins, viewsets

from resolwe.flow.filters import ProcessFilter
from resolwe.flow.models import Process
from resolwe.flow.serializers import ProcessSerializer
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin
from resolwe.permissions.models import PermissionModel

from .mixins import ResolweCheckSlugMixin, ResolweCreateModelMixin


class ProcessViewSet(
    ResolweCreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    ResolwePermissionsMixin,
    ResolweCheckSlugMixin,
    viewsets.GenericViewSet,
):
    """API view for :class:`Process` objects."""

    qs_permission_model = PermissionModel.objects.select_related("user", "group")
    queryset = Process.objects.all().select_related("contributor")
    serializer_class = ProcessSerializer
    permission_classes = (get_permissions_class(),)
    filterset_class = ProcessFilter
    ordering_fields = ("id", "created", "modified", "name", "version")
    ordering = ("id",)

    def create(self, request, *args, **kwargs):
        """Only superusers can create new processes."""
        if not request.user.is_superuser:
            raise exceptions.NotFound

        return super().create(request, *args, **kwargs)

    def get_queryset(self):
        """Prefetch permissions for current user."""
        return self.prefetch_current_user_permissions(self.queryset)



# pylint: disable=missing-docstring
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.urls import reverse

from rest_framework.test import APITestCase

from resolwe.flow.models import Collection, Data, Entity, Process
from resolwe.permissions.models import Permission


class ProcessOrderingTest(APITestCase):
    def setUp(self):
        super().setUp()

        user_model = get_user_model()
        user = user_model.objects.create(username="user")

        self.proc_1 = Process.objects.create(
            name="My process", contributor=user, version=1
        )
        self.proc_2 = Process.objects.create(
            name="My process", contributor=user, version=2
        )

        self.proc_1.set_permission(Permission.VIEW, AnonymousUser())
        self.proc_2.set_permission(Permission.VIEW, AnonymousUser())

        self.url = reverse("resolwe-api:process-list")

    def test_ordering_version(self):
        response = self.client.get(self.url, {"ordering": "version"}, format="json")
        self.assertEqual(response.data[0]["id"], self.proc_1.id)
        self.assertEqual(response.data[1]["id"], self.proc_2.id)

        response = self.client.get(self.url, {"ordering": "-version"}, format="json")
        self.assertEqual(response.data[0]["id"], self.proc_2.id)
        self.assertEqual(response.data[1]["id"], self.proc_1.id)


class CollectionOrderingTest(APITestCase):
    def setUp(self):
        super().setUp()

        user_model = get_user_model()
        self.user_1 = user_model.objects.create(
            username="user_1", first_name="Zion", last_name="Zucchini"
        )
        self.user_2 = user_model.objects.create(
            username="user_2", first_name="Zack", last_name="Zucchini"
        )
        self.user_3 = user_model.objects.create(
            username="user_3", first_name="Adam", last_name="Angus"
        )

        self.collection_1 = Collection.objects.create(
            name="Collection 1", contributor=self.user_1
        )
        self.collection_2 = Collection.objects.create(
            name="Collection 2", contributor=self.user_2
        )
        self.collection_3 = Collection.objects.create(
            name="Collection 3", contributor=self.user_3
        )

        self.collection_1.set_permission(Permission.VIEW, AnonymousUser())
        self.collection_2.set_permission(Permission.VIEW, AnonymousUser())
        self.collection_3.set_permission(Permission.VIEW, AnonymousUser())

        self.url = reverse("resolwe-api:collection-list")

    def test_ordering(self):
        response = self.client.get(
            self.url, {"ordering": "contributor__last_name,contributor__first_name"}
        )
        self.assertEqual(
            response.data[0]["contributor"]["first_name"], self.user_3.first_name
        )
        self.assertEqual(
            response.data[1]["contributor"]["first_name"], self.user_2.first_name
        )
        self.assertEqual(
            response.data[2]["contributor"]["first_name"], self.user_1.first_name
        )

        response = self.client.get(
            self.url, {"ordering": "-contributor__last_name,-contributor__first_name"}
        )
        self.assertEqual(
            response.data[0]["contributor"]["first_name"], self.user_1.first_name
        )
        self.assertEqual(
            response.data[1]["contributor"]["first_name"], self.user_2.first_name
        )
        self.assertEqual(
            response.data[2]["contributor"]["first_name"], self.user_3.first_name
        )


class EntityOrderingTest(APITestCase):
    def setUp(self):
        super().setUp()

        user_model = get_user_model()
        self.user_1 = user_model.objects.create(
            username="user_1", first_name="Zion", last_name="Zucchini"
        )
        self.user_2 = user_model.objects.create(
            username="user_2", first_name="Zack", last_name="Zucchini"
        )
        self.user_3 = user_model.objects.create(
            username="user_3", first_name="Adam", last_name="Angus"
        )

        self.entity_1 = Entity.objects.create(name="Entity 1", contributor=self.user_1)
        self.entity_2 = Entity.objects.create(name="Entity 2", contributor=self.user_2)
        self.entity_3 = Entity.objects.create(name="Entity 3", contributor=self.user_3)

        self.entity_1.set_permission(Permission.VIEW, AnonymousUser())
        self.entity_2.set_permission(Permission.VIEW, AnonymousUser())
        self.entity_3.set_permission(Permission.VIEW, AnonymousUser())

        self.url = reverse("resolwe-api:entity-list")

    def test_ordering(self):
        response = self.client.get(
            self.url, {"ordering": "contributor__last_name,contributor__first_name"}
        )
        self.assertEqual(
            response.data[0]["contributor"]["first_name"], self.user_3.first_name
        )
        self.assertEqual(
            response.data[1]["contributor"]["first_name"], self.user_2.first_name
        )
        self.assertEqual(
            response.data[2]["contributor"]["first_name"], self.user_1.first_name
        )

        response = self.client.get(
            self.url, {"ordering": "-contributor__last_name,-contributor__first_name"}
        )
        self.assertEqual(
            response.data[0]["contributor"]["first_name"], self.user_1.first_name
        )
        self.assertEqual(
            response.data[1]["contributor"]["first_name"], self.user_2.first_name
        )
        self.assertEqual(
            response.data[2]["contributor"]["first_name"], self.user_3.first_name
        )


class DataOrderingTest(APITestCase):
    def setUp(self):
        super().setUp()

        user_model = get_user_model()
        self.user_1 = user_model.objects.create(
            username="user_1", first_name="Zion", last_name="Zucchini"
        )
        self.user_2 = user_model.objects.create(
            username="user_2", first_name="Zack", last_name="Zucchini"
        )
        self.user_3 = user_model.objects.create(
            username="user_3", first_name="Adam", last_name="Angus"
        )

        process = Process.objects.create(name="My process", contributor=self.user_1)
        self.data_1 = Data.objects.create(
            name="Data 1", contributor=self.user_1, process=process
        )
        self.data_2 = Data.objects.create(
            name="Data 2", contributor=self.user_2, process=process
        )
        self.data_3 = Data.objects.create(
            name="Data 3", contributor=self.user_3, process=process
        )

        self.data_1.set_permission(Permission.VIEW, AnonymousUser())
        self.data_2.set_permission(Permission.VIEW, AnonymousUser())
        self.data_3.set_permission(Permission.VIEW, AnonymousUser())

        self.url = reverse("resolwe-api:data-list")

    def test_ordering(self):
        response = self.client.get(
            self.url, {"ordering": "contributor__last_name,contributor__first_name"}
        )
        self.assertEqual(
            response.data[0]["contributor"]["first_name"], self.user_3.first_name
        )
        self.assertEqual(
            response.data[1]["contributor"]["first_name"], self.user_2.first_name
        )
        self.assertEqual(
            response.data[2]["contributor"]["first_name"], self.user_1.first_name
        )

        response = self.client.get(
            self.url, {"ordering": "-contributor__last_name,-contributor__first_name"}
        )
        self.assertEqual(
            response.data[0]["contributor"]["first_name"], self.user_1.first_name
        )
        self.assertEqual(
            response.data[1]["contributor"]["first_name"], self.user_2.first_name
        )
        self.assertEqual(
            response.data[2]["contributor"]["first_name"], self.user_3.first_name
        )



"""Permissions functions used in Resolwe Viewsets."""
from django.conf import settings
from django.db import models, transaction

from rest_framework import exceptions, serializers, status
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response

from resolwe.auditlog.auditmanager import AuditManager
from resolwe.permissions.models import Permission
from resolwe.permissions.shortcuts import get_object_perms

from .utils import (
    check_owner_permission,
    check_public_permissions,
    check_user_permissions,
    update_permission,
)


class CurrentUserPermissionsSerializer(serializers.Serializer):
    """Current user permissions serializer."""

    id = serializers.IntegerField()
    type = serializers.CharField(max_length=50)
    name = serializers.CharField(max_length=100)
    permissions = serializers.ListField(child=serializers.CharField(max_length=30))


class ResolwePermissionsMixin:
    """Mixin to support managing `Resolwe` objects' permissions."""

    def prefetch_current_user_permissions(self, queryset: models.QuerySet):
        """Prefetch permissions for the current user."""
        user = self.request.user
        filters = models.Q(user__username=settings.ANONYMOUS_USER_NAME)
        if not user.is_anonymous:
            filters |= models.Q(user=user) | models.Q(group__in=user.groups.all())

        qs_permission_model = self.qs_permission_model.filter(filters)
        return queryset.prefetch_related(
            models.Prefetch(
                "permission_group__permissions", queryset=qs_permission_model
            )
        )

    def get_serializer_class(self):
        """Augment base serializer class.

        Include permissions information with objects.

        """
        base_class = super().get_serializer_class()

        class SerializerWithPermissions(base_class):
            """Augment serializer class."""

            def get_fields(serializer_self):
                """Return serializer's fields."""
                fields = super().get_fields()
                fields["current_user_permissions"] = CurrentUserPermissionsSerializer(
                    read_only=True
                )
                return fields

            def to_representation(serializer_self, instance: models.Model):
                """Object serializer."""
                data = super().to_representation(instance)
                if (
                    "fields" not in self.request.query_params
                    or "current_user_permissions" in self.request.query_params["fields"]
                ):
                    data["current_user_permissions"] = get_object_perms(
                        instance, self.request.user, True
                    )
                return data

        return SerializerWithPermissions

    @action(
        detail=True,
        methods=["get", "post"],
        url_path="permissions",
        url_name="permissions",
    )
    def detail_permissions(self, request: Request, pk=None) -> Response:
        """Get or set permissions API endpoint."""
        # The object is taken from the queryset on the view for which
        # permissions are prefetched for the current user only.
        # This implies that obj.permission_group.permissions returns
        # permissions for the current user.
        # To get all the permissions we have to perform a refresh from the
        # database. This must only be done if user has share permission on the
        # given object otherwise only his permissions must be returned.
        obj = self.get_object()

        if obj.has_permission(Permission.SHARE, request.user):
            obj.refresh_from_db()

        audit_manager = AuditManager.global_instance()

        if request.method == "POST":
            audit_manager.log_message("Permissions updated: %s", request.data)
            allow_owner = obj.is_owner(request.user) or request.user.is_superuser
            check_owner_permission(request.data, allow_owner, obj)
            check_public_permissions(request.data)
            check_user_permissions(request.data, request.user.pk)

            with transaction.atomic():
                update_permission(obj, request.data)
                owner_count = obj.permission_group.permissions.filter(
                    value=Permission.OWNER.value, user__isnull=False
                ).count()

                if not owner_count:
                    raise exceptions.ParseError("Object must have at least one owner.")
        else:
            audit_manager.log_message("Permissions read: %s", request.data)

        return Response(get_object_perms(obj))

    @action(
        detail=False,
        methods=["get", "post"],
        url_path="permissions",
        url_name="permissions",
    )
    def list_permissions(self, request: Request) -> Response:
        """Batch get or set permissions API endpoint."""
        # TODO: Implement batch get/set permissions
        return Response(status=status.HTTP_501_NOT_IMPLEMENTED)



# Generated by Django 3.1.7 on 2021-10-12 10:39

import django.db.models.deletion
import django.db.models.manager
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("flow", "0001_squashed_0043_full_text_search"),
    ]

    operations = [
        migrations.CreateModel(
            name="FileStorage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name="StorageLocation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("url", models.CharField(max_length=60)),
                ("connector_name", models.CharField(max_length=30)),
                ("last_update", models.DateTimeField(auto_now=True)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("PR", "Preparing"),
                            ("UP", "Uploading"),
                            ("OK", "Done"),
                            ("DE", "Deleting"),
                        ],
                        default="PR",
                        max_length=2,
                    ),
                ),
                (
                    "file_storage",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="storage_locations",
                        to="storage.filestorage",
                    ),
                ),
            ],
            options={
                "unique_together": {("url", "connector_name")},
            },
            managers=[
                ("all_objects", django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name="ReferencedPath",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("path", models.TextField(db_index=True)),
                ("size", models.BigIntegerField(default=-1)),
                ("md5", models.CharField(max_length=32)),
                ("crc32c", models.CharField(max_length=8)),
                ("awss3etag", models.CharField(max_length=50)),
                ("chunk_size", models.IntegerField(default=8388608)),
                (
                    "storage_locations",
                    models.ManyToManyField(
                        related_name="files", to="storage.StorageLocation"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AccessLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("started", models.DateTimeField(auto_now_add=True)),
                ("finished", models.DateTimeField(blank=True, null=True)),
                ("reason", models.CharField(max_length=120)),
                (
                    "cause",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="access_logs",
                        to="flow.data",
                    ),
                ),
                (
                    "storage_location",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="access_logs",
                        to="storage.storagelocation",
                    ),
                ),
            ],
        ),
    ]



"""Resolwe models hydrate utils."""
import copy
import os
import re
from pathlib import Path

from django.core.exceptions import ValidationError

from resolwe.flow.utils import iterate_fields


def _hydrate_values(output, output_schema, data):
    """Hydrate basic:file and basic:json values.

    Find fields with basic:file type and assign a full path to the file.
    Find fields with basic:json type and assign a JSON object from storage.

    """

    def hydrate_path(file_name):
        """Hydrate file paths."""
        from resolwe.flow.managers import manager

        class HydratedPath(str):
            """String wrapper, which also stores the original filename."""

            __slots__ = ("data_id", "file_name")

            def __new__(cls, value=""):
                """Initialize hydrated path."""
                hydrated = str.__new__(cls, value)
                hydrated.data_id = data.id
                hydrated.file_name = file_name
                return hydrated

        return HydratedPath(manager.get_executor().resolve_data_path(data, file_name))

    def hydrate_storage(storage_id):
        """Hydrate storage fields."""
        from ..storage import LazyStorageJSON  # Prevent circular import.

        return LazyStorageJSON(pk=storage_id)

    for field_schema, fields in iterate_fields(output, output_schema):
        name = field_schema["name"]
        value = fields[name]
        if "type" in field_schema:
            if field_schema["type"].startswith("basic:file:"):
                value["file"] = hydrate_path(value["file"])
                value["refs"] = [hydrate_path(ref) for ref in value.get("refs", [])]

            elif field_schema["type"].startswith("list:basic:file:"):
                for obj in value:
                    obj["file"] = hydrate_path(obj["file"])
                    obj["refs"] = [hydrate_path(ref) for ref in obj.get("refs", [])]

            if field_schema["type"].startswith("basic:dir:"):
                value["dir"] = hydrate_path(value["dir"])
                value["refs"] = [hydrate_path(ref) for ref in value.get("refs", [])]

            elif field_schema["type"].startswith("list:basic:dir:"):
                for obj in value:
                    obj["dir"] = hydrate_path(obj["dir"])
                    obj["refs"] = [hydrate_path(ref) for ref in obj.get("refs", [])]

            elif field_schema["type"].startswith("basic:json:"):
                fields[name] = hydrate_storage(value)

            elif field_schema["type"].startswith("list:basic:json:"):
                fields[name] = [hydrate_storage(storage_id) for storage_id in value]


def hydrate_input_references(input_, input_schema, hydrate_values=True):
    """Hydrate ``input_`` with linked data.

    Find fields with complex data:<...> types in ``input_``.
    Assign an output of corresponding data object to those fields.

    """
    from ..data import Data  # prevent circular import

    for field_schema, fields in iterate_fields(input_, input_schema):
        name = field_schema["name"]
        value = fields[name]
        if "type" in field_schema:
            if field_schema["type"].startswith("data:"):
                if value is None:
                    continue

                try:
                    data = Data.objects.get(id=value)
                except Data.DoesNotExist:
                    fields[name] = {}
                    continue

                output = copy.deepcopy(data.output)
                hydrate_input_references(output, data.process.output_schema)
                if hydrate_values:
                    _hydrate_values(output, data.process.output_schema, data)
                output["__id"] = data.id
                output["__type"] = data.process.type
                output["__descriptor"] = data.descriptor
                output["__name"] = getattr(data, "name", None)
                output["__entity_id"] = getattr(data.entity, "id", None)
                output["__entity_name"] = getattr(data.entity, "name", None)
                output["__output_schema"] = data.process.output_schema

                fields[name] = output

            elif field_schema["type"].startswith("list:data:"):
                outputs = []
                for val in value:
                    if val is None:
                        continue

                    try:
                        data = Data.objects.get(id=val)
                    except Data.DoesNotExist:
                        outputs.append({})
                        continue

                    output = copy.deepcopy(data.output)
                    hydrate_input_references(output, data.process.output_schema)
                    if hydrate_values:
                        _hydrate_values(output, data.process.output_schema, data)

                    output["__id"] = data.id
                    output["__type"] = data.process.type
                    output["__descriptor"] = data.descriptor
                    output["__name"] = getattr(data, "name", None)
                    output["__entity_id"] = getattr(data.entity, "id", None)
                    output["__entity_name"] = getattr(data.entity, "name", None)
                    output["__output_schema"] = data.process.output_schema

                    outputs.append(output)

                fields[name] = outputs


def hydrate_input_uploads(input_, input_schema, hydrate_values=True):
    """Hydrate input basic:upload types with upload location.

    Find basic:upload fields in input.
    Add the upload location for relative paths.

    """
    from resolwe.flow.managers import manager

    files = []
    for field_schema, fields in iterate_fields(input_, input_schema):
        name = field_schema["name"]
        value = fields[name]
        if "type" in field_schema:
            if field_schema["type"] == "basic:file:":
                files.append(value)

            elif field_schema["type"] == "list:basic:file:":
                files.extend(value)

    urlregex = re.compile(
        r"^(https?|ftp|s3)://[-A-Za-z0-9\+&@#/%?=~_|!:,.;]*[-A-Za-z0-9\+&@#/%=~_|]"
    )
    for value in files:
        if "file_temp" in value:
            if isinstance(value["file_temp"], str):
                # If file_temp not url, hydrate path.
                if not urlregex.search(value["file_temp"]):
                    value["file_temp"] = manager.get_executor().resolve_upload_path(
                        value["file_temp"]
                    )
            else:
                # Something very strange happened.
                value["file_temp"] = "Invalid value for file_temp in DB"


def hydrate_size(data, force=False):
    """Add file and dir sizes.

    Add sizes to ``basic:file:``, ``list:basic:file``, ``basic:dir:``
    and ``list:basic:dir:`` fields.

    ``force`` parameter is used to recompute file sizes also on objects
    that already have these values, e.g. in migrations.
    """
    from ..data import Data  # prevent circular import

    def get_dir_size(path):
        """Get directory size."""
        return sum(
            file_.stat().st_size for file_ in Path(path).rglob("*") if file_.is_file()
        )

    def get_refs_size(obj, obj_path):
        """Calculate size of all references of ``obj``.

        :param dict obj: Data object's output field (of type file/dir).
        :param Path obj_path: Path to ``obj``.
        """
        total_size = 0
        for ref in obj.get("refs", []):
            ref_path = data.location.get_path(filename=ref)
            if ref_path in os.fspath(obj_path):
                # It is a common case that ``obj['file']`` is also contained in
                # one of obj['ref']. In that case, we need to make sure that it's
                # size is not counted twice:
                continue
            ref_path: Path = Path(ref_path)
            if ref_path.is_file():
                total_size += ref_path.stat().st_size
            elif ref_path.is_dir():
                total_size += get_dir_size(ref_path)

        return total_size

    def add_file_size(obj):
        """Add file size to the basic:file field."""
        if (
            data.status in [Data.STATUS_DONE, Data.STATUS_ERROR]
            and "size" in obj
            and not force
        ):
            return

        path = Path(data.location.get_path(filename=obj["file"]))
        if not path.is_file():
            raise ValidationError("Referenced file does not exist ({})".format(path))

        obj["size"] = path.stat().st_size
        obj["total_size"] = obj["size"] + get_refs_size(obj, path)

    def add_dir_size(obj):
        """Add directory size to the basic:dir field."""
        if (
            data.status in [Data.STATUS_DONE, Data.STATUS_ERROR]
            and "size" in obj
            and not force
        ):
            return

        path = Path(data.location.get_path(filename=obj["dir"]))
        if not path.is_dir():
            raise ValidationError("Referenced dir does not exist ({})".format(path))

        obj["size"] = get_dir_size(path)
        obj["total_size"] = obj["size"] + get_refs_size(obj, path)

    data_size = 0
    for field_schema, fields in iterate_fields(data.output, data.process.output_schema):
        name = field_schema["name"]
        value = fields[name]
        if "type" in field_schema:
            if field_schema["type"].startswith("basic:file:"):
                add_file_size(value)
                data_size += value.get("total_size", 0)
            elif field_schema["type"].startswith("list:basic:file:"):
                for obj in value:
                    add_file_size(obj)
                    data_size += obj.get("total_size", 0)
            elif field_schema["type"].startswith("basic:dir:"):
                add_dir_size(value)
                data_size += value.get("total_size", 0)
            elif field_schema["type"].startswith("list:basic:dir:"):
                for obj in value:
                    add_dir_size(obj)
                    data_size += obj.get("total_size", 0)

    data.size = data_size
